#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2012 Liricus SRL <info@liricus.com.ar>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Fra


#===============================================================================
# DOCS
#===============================================================================

"""Loads the Flask application for the socialframe server.

Instances the mail client, database and authentication client.

"""


#===============================================================================
# IMPORTS
#===============================================================================

import sys
import traceback
import functools
import datetime
import re
import time
import threading

import flask
import flask_mail 
from flaskext import uploads
from flask_peewee import db

from werkzeug import routing

from socialframe_server import settings


#===============================================================================
# CONSTANTS
#===============================================================================

APP = flask.Flask("socialframe_server")
APP.config.from_object(settings)

APP_DB = db.Database(APP)

APP_MAIL = flask_mail.Mail(APP)

APP_AUTH = None # this is setted when models is imported


#===============================================================================
# VALIDATORS 
#===============================================================================

errors = []
if "original" in settings.IMAGE_SIZES:
    errors.append("'original' image size name is not allowed")
    
if errors:
    msg = "\n".join(errors)
    raise ValueError(msg)
    
    
#===============================================================================
# CONFIGS
#===============================================================================

if settings.MAX_UPLOAD_SIZE > 0:
    uploads.patch_request_class(APP, settings.MAX_UPLOAD_SIZE)
    

#===============================================================================
# CONVERTERS
#===============================================================================

class ListConverter(routing.BaseConverter):
    def to_python(self, value):
        l = []
        for v in value.split(','):
            if v == "null":
                v = None
            l.append(v)
        return l
    def to_url(self, values):
        l = []
        for v in values:
            if v is None:
                v = "null"
            l.append(routing.BaseConverter.to_url(v))
        return ','.join(l)
APP.url_map.converters['list'] = ListConverter


class DictConverter(routing.BaseConverter):
    def to_python(self, value):
        d = {}
        for part in value.split(","):
            k, v = part.split(":")
            if v == "null":
                v = None
            elif v == "true":
                v = True
            elif v == "false":
                v = False
            d[k] = v
        return d 
    def to_url(self, values):
        parts = []
        for k, v in values.items():
            if v is None:
                v = "null"
            parts.append(routing.BaseConverter.to_url(k + ":" + v))
        return ','.join(parts)
APP.url_map.converters['dict'] = DictConverter


class NullListConverter(routing.BaseConverter):
    def to_python(self, value):
        if value != "null":
            l = []
            for v in value.split(','):
                if v == "null":
                    v = None
                l.append(v)
            return l
        else:
            return None
    def to_url(self, values):
        if values != None:
            l = []
            for v in values:
                if v is None:
                    v = "null"
                l.append(routing.BaseConverter.to_url(v))
            return ','.join(l)
        else:
            return "null"
APP.url_map.converters['nulllist'] = NullListConverter


class NullConverter(routing.BaseConverter):
    def to_python(self, value):
        return None if value == "null" else value
    def to_url(self, value):
        return routing.BaseConverter.to_url(
            "null" if value == None else value
        )
APP.url_map.converters['null'] = NullConverter


class DateConverter(routing.BaseConverter):
    def to_python(self, value):
        return datetime.datetime.strptime(value, "%Y%m%d")
    def to_url(self, value):
        return value.strftime("%Y%m%d")
APP.url_map.converters['date'] = DateConverter


class BoolConverter(routing.BaseConverter):
    def to_python(self, value):
        if value == "true":
            return True
        elif value == "false":
            return False
        flask.abort(400)
    def to_url(self, value):
        return routing.BaseConverter.to_url(
            "true" if value else "false"
        )
APP.url_map.converters['bool'] = BoolConverter


class EmailConverter(routing.BaseConverter):
    RX_MAIL = re.compile("^[a-zA-Z0-9._%-]+@[a-zA-Z0-9._%-]+.[a-zA-Z]{2,6}$")
    def to_python(self, value):
        if not self.RX_MAIL.match(value):
            flask.abort(400)
        return value
    def to_url(self, value):
        return value
APP.url_map.converters['email'] = EmailConverter
    


#===============================================================================
# ERROR HANDLERS
#===============================================================================

@APP.errorhandler(404)    
@APP.errorhandler(403)
@APP.errorhandler(500)
@APP.errorhandler(400)
@APP.errorhandler(Exception)
def errorhandler_json(ex):
    """Returns error code and information in JSON format.
    
        **Errors handled**
        
        The Errors are ``400``, ``403``, ``404``, ``500``.
    
    """
    
    code = getattr(ex, "code", None)
    if isinstance(code, basestring) and code.isdigit():
        code = int(code)
    elif not isinstance(code, int):
        code = 500
    
    data = {
        "error":True,
        "code": code,
        "msg":ex.message,
    }
    if settings.DEBUG:
        traceinfo = "".join(traceback.format_exception(*sys.exc_info()))
        data["traceback"] = traceinfo
    return json_data(**data)


#===============================================================================
# FUNCTIONS
#===============================================================================

def absurl(*args):
    """Creates an absolute url mapped with the given arguments and returns it.
    
    """
    
    parts = [
        "http://{host}:{port}".format(host=settings.HOST, port=settings.PORT)
    ]
    for arg in args:
        if arg.startswith("/"):
            arg = arg[1:]
        parts.append(arg)
    return "/".join(parts)


def json_data(**args):
    data = dict(args)
    data["action"] = flask.request.endpoint
    if "error" not in data:
        data["error"] = False
    data["code"] = data.get("code", 200)
    resp = flask.jsonify(**data)
    resp.status_code = data["code"]
    return resp


_unique = int(time.time())
def unique():
    """Return a unique code for a statting with sistem time when
    the project is started
    
    """
    global _unique
    _unique += 1
    return _unique


_jobs = set()
def run_jobs():
    global _jobs
    for func in _jobs:
        print "Job '{}' run...".format(func.__name__)
        with APP_DB.database.transaction():
            func()


def get_from_dict(d, keys, default=settings.DEFAULT_LANG):
    """Returns the first ocurrence of key from a given dict otherwise return
    the defaul key
    
    """
    for k in keys:
        if k in d:
            return d[k]
    return d[default]


#===============================================================================
# DECORATORS
#===============================================================================

def login_required(func):
    """Checks if user is logged, then original function ``func`` proceeds.
    If not throws ``Error 403``.
    
    """
    
    @functools.wraps(func)
    def _wrap(*args, **kwargs):
        if APP_AUTH.get_logged_in_user():
            return func(*args, **kwargs)
        else:
            flask.abort(403)
            
    return _wrap


def transactional(func):
    @functools.wraps(func)
    def _wrap(*args, **kwargs):
        with APP_DB.database.transaction():
            return func(*args, **kwargs)
    return _wrap


def job(func):
    global _jobs
    _jobs.add(func)
    return func
        

#===============================================================================
# MAIN
#===============================================================================

if __name__ == "__main__":
    print(__doc__)
